
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _GNU_SOURCE
#include <fcntl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <getopt.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>
#include <errno.h>
#include <poll.h>

#include <initng.h>
#include <initng-paths.h>

#include "libngeclient.h"

const char * ngeclient_error = NULL;

/* open socket, specified path */
int ngeclient_open_socket(nge_event * event, const char * path)
{
    int len;
    struct sockaddr_un sockname;

    assert(event);
    assert(path);

    /* check that socket is not open allredy */
    if(event->sock > 1)
    {
	ngeclient_error="Socket allredy open.";
	/*fprintf(stderr, "Socket allredy open.\n");*/
	return(-1);
    }
	
    event->sock=-1;
    event->read_buffer=NULL;
    event->read_buffer_len=0;

    /* Create the socket. */
    event->sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if (event->sock < 0)
    {
	ngeclient_error="Failed to init socket.";
        /*fprintf(stderr, "Failed to init socket.");*/
        return (-1);
    }

    /* Bind a name to the socket. */
    sockname.sun_family = AF_UNIX;
    strcpy(sockname.sun_path, path);
    len = strlen(path) + sizeof(sockname.sun_family);

    if (connect(event->sock, (struct sockaddr *) &sockname, len) < 0)
    {
        close(event->sock);
	ngeclient_error="Error connecting to socket";
        /*fprintf(stderr, "Error connecting to socket\n");*/
        return (-1);
    }
    
    
    /* Put it not to block, waiting for more data on rscv */
    {
	int cur = fcntl(event->sock, F_GETFL, 0);
	fcntl(event->sock, F_SETFL, cur | O_NONBLOCK);
    }
    
    /* return happily */
    return (event->sock);
}

/* This will try to free the buffer compleatly */
void ngeclient_free_buffer(nge_event * event)
{
    if(event->read_buffer)
    {
	free(event->read_buffer);
	event->read_buffer=NULL;
	event->read_buffer_len=0;
    }
}

/* close pipes */
void ngeclient_close_socket(nge_event * event)
{
    assert(event);
    
    /* if socket is open, close it */
    if (event->sock > 0)
    {
        close(event->sock);
    }

    /* set socket marked closed */
    event->sock = -1;
    
    /* free read_buffer */
    if(event->read_buffer)
    {
	free(event->read_buffer);
	event->read_buffer=NULL;
    }
}




/*
 * read_and_fill_buffer()
 * This function fetches data on the input fd sock (globally defined)
 * it fills read_buffer, that is reallocad to fill data.
 */
int ngeclient_read_and_fill_buffer(nge_event * event)
{
    int totally_got = 0;
    int chars_recv = 1; /* must be one so the while loop will be run */
    
    assert(event);
    assert(event->sock > 0);
    
    /* while we got data from recv, (or the initziating time ) */
    while(chars_recv > 0)
    {
    
	/*printf("reallocating to %i\n", event->read_buffer_len + 101 );*/
	/* Allocate more room for input */
	event->read_buffer = realloc(event->read_buffer, (event->read_buffer_len + 101) * sizeof(char));
	
	/* Make its null terminated, specially if event->read_buffer is created from NULL */
	event->read_buffer[event->read_buffer_len]='\0';

	/* get 100 chars */
	chars_recv = recv(event->sock, &event->read_buffer[event->read_buffer_len], 100, 0);
	
	/* null the end */
	if(chars_recv > 0)
	{
	    event->read_buffer_len+=chars_recv;
	    event->read_buffer[event->read_buffer_len]='\0';
	    totally_got += chars_recv;
	}
    }
    
    /* return chars got */
    return(totally_got);
}

/*
 * poll_for_input(seconds)
 * Will wait for (seconds) and if there is data incomming.
 * event->read_buffer is filled, and this will return the chars read.
 */
int ngeclient_poll_for_input(nge_event * event, int seconds)
{
    int ret_poll;
    struct pollfd fds[1];
    
    assert(event);
    
    /* fille in the poll */
    fds[0].fd = event->sock;
    fds[0].events = POLLIN;
    
    /* now wait for data to apeare */
    ret_poll = poll(fds, 1, seconds);
    
    /* if it was any data */
    if(ret_poll > 0)
	/* go fetch it */
	return(ngeclient_read_and_fill_buffer(event));

    /* return the 0 or -1 poll generates */
    return(ret_poll);
}

/*
 * This function cuts event->read_buffer from
 * beginning and then reallocates to save memory.
 */
void ngeclient_cut_buffert(nge_event * event, int chars)
{
    assert(event);

    /* dont cut any then */
    if(chars == 0 || !event->read_buffer || event->read_buffer_len<1)
	return;

    /* make sure we are not trying to cut more then we have */
    if(chars > event->read_buffer_len)
	chars = event->read_buffer_len;
    
    /* get lenght of buffer that is left, after chars is cut */
    event->read_buffer_len -= chars;
    
    /* move the content in front, to the beginning */
    memmove(event->read_buffer, &event->read_buffer[chars], event->read_buffer_len * sizeof(char));
    
    /* reallocate now, when theres not that mutch data anymore */
    /*printf("reallocating to %i\n", event->read_buffer_len);*/
    event->read_buffer=realloc(event->read_buffer, (event->read_buffer_len + 1) * sizeof(char));

    /* make sure its nulled at end */
    event->read_buffer[event->read_buffer_len] = '\0';
}


/*
 * ngeclient_get_content(tag)
 * If tag is "<specialtag>This is some data.</specialtag>"
 * This returns a new allocated strng "This is some data."
 */
 
static char *ngeclient_get_content(const char *tag)
{
    int len = 0;
    char * point = (char *) tag;
    assert(tag);
    
    /*
     * Maks sure we stand here:
     * "<specialtag>This is some data.</specialtag>"
     *  |
     */
    if(point[0]!='<')
    {
	ngeclient_error="ngeclient_get_content failed. Bad xml.";
	return(NULL);
    }
    
    /* step to the end of tag */
    while(point[0] && point[0]!='>')
	point++;
    
    /* make sure we got there */
    if(point[0]!='>')
    {
	ngeclient_error="ngeclient_get_content failed. Bad xml.";
	return(NULL);
    }
    point++;
    
    /*
     * Maks sure we stand here:
     * "<specialtag>This is some data.</specialtag>"
     *              |
     */

    while(point[len] && point[len]!='<')
	len++;
    
    /* make sure we are standing on start on next tag. */
    if(point[len] != '<')
    {
	ngeclient_error="ngeclient_get_content failed. Bad xml.";
	return(NULL);
    }

    /*
     * Maks sure we stand here:
     * "<specialtag>This is some data.</specialtag>"
     *              |                 ^
     */
    if(len>0)
	return(strndup(point, len));

    /* else return ZERO */
    return(NULL);
}

/*
 * ngeclient_get_option(tag, name);
 * Say we have a tag "<connect initng_version="initng-123">"
 * Then calling ngeclient_get_option(tag, "initng_version")
 * will return a string "initng-123"
 */
static char *ngeclient_get_option(const char *tag, const char * name)
{
    assert(tag);
    assert(name);
    char *point=(char *)tag;
    int name_len=strlen(name);
    
    /* search for the name in tag */
    while((point=strstr(point, name)))
    {
	int len=0;
    
	/* Standing here:
	 * "<connect initng_version="initng-123">"
	 *           |
	 */
     
	point += name_len;
	/* Standing here:
	 * "<connect initng_version="initng-123">"
	 *                         |
	 */
    
	/* skip spaces */
	while(point[0] && point[0]==' ' && point[0]=='\t')
	    point++;

	/* must be syntax */
	if(point[0]!='=')
	    continue;
        point++;

        /* skip spaces */
        while(point[0] && point[0]==' ' && point[0]=='\t')
	    point++;

        /* Standing here:
	 * "<connect initng_version="initng-123">"
         *                          |
         */

	/* must be syntax */
        if(point[0]!='"')
	    continue;
        point++;
	
	/* Standing here:
         * "<connect initng_version="initng-123">"
         *                           |
         */
    
        while(point[len] && point[len]!='"')
	    len++;

        /* Standing here: | To here: ^
         * "<connect initng_version="initng-123">"
         *                           |         ^
         */

        /* return that string */
        return(strndup(point, len));
    }
    
    ngeclient_error="Coud not find option in tag.";
    return(NULL);
}

/*
 * this is a quick hack, for getting an int, outpuf an
 * ngeclient_get_option
 */
static int ngeclient_get_int(const char *tag, const char * name)
{
    int value;
    char *string;
    assert(tag);
    assert(name);
    string=ngeclient_get_option(tag, name);
    if(!string) return(-1);
    value=atoi(string);
    free(string);
    return(value);
}


/* called on a <event type="service_state_change"  */
static void ngeclient_handle_service_state_change(nge_event * event, char *tag, int chars)
{
    char *service = ngeclient_get_option(tag, "service");
    e_is is = (e_is) ngeclient_get_int(tag, "is");
    char *state = ngeclient_get_option(tag, "state");
    
    if(event->service_state_change)
	(*event->service_state_change)(event, service, is, state);
	
    /*printf("service_state_change \"%s\", %i chars.\n %s, %i %s \n", tag, chars, service, is , state);*/

    free(service);
    free(state);
}

/* called on a <event type="system_state_change"  */
static void ngeclient_handle_system_state_change(nge_event * event, char *tag, int chars)
{
    h_sys_state system_state = (h_sys_state) ngeclient_get_int(tag, "system_state");
    
    if(event->system_state_change)
	(*event->system_state_change)(event, system_state);

    /*printf("event-tag \"%s\", %i chars.\n", tag, chars);*/

}

/* called on a <event type="initial_service_state"  */
static void ngeclient_handle_initial_service_state(nge_event * event, char *tag, int chars)
{
    char *service = ngeclient_get_option(tag, "service");
    e_is is = (e_is) ngeclient_get_int(tag, "is");
    char *state = ngeclient_get_option(tag, "state");
    
    if(event->initial_service_state)
	(*event->initial_service_state)(event, service, is, state);
	
    /*printf("service_state_change \"%s\", %i chars.\n %s, %i %s \n", tag, chars, service, is , state);*/

    free(service);
    free(state);
}

/* called on a <event type="initial_system_state"  */
static void ngeclient_handle_initial_system_state(nge_event * event, char *tag, int chars)
{
    h_sys_state system_state = (h_sys_state) ngeclient_get_int(tag, "system_state");
    
    if(event->initial_system_state)
	(*event->initial_system_state)(event, system_state);

    /*printf("event-tag \"%s\", %i chars.\n", tag, chars);*/

}

/* called on a <event type="service_output"  */
static void ngeclient_handle_service_output(nge_event * event, char *tag, int chars)
{
    char * service = ngeclient_get_option(tag, "service");
    char * process = ngeclient_get_option(tag, "process");
    char * output = ngeclient_get_content(tag);
    /*printf("Service \"%s\" process \"%s\" outputed:\n%s\n", service, process, output);*/
    
    if(event->service_output)
	(*event->service_output)(event, service, process, output);

    free(service);
    free(process);
    free(output);
}

/* called on a <event type="err_msg"  */
static void ngeclient_handle_err_msg(nge_event * event, char *tag, int chars)
{
    e_mt mt = (e_mt) ngeclient_get_int(tag, "mt");
    char *file = ngeclient_get_option(tag, "file");
    char *func = ngeclient_get_option(tag, "func");
    int line = ngeclient_get_int(tag, "line");
    char *message = ngeclient_get_content(tag);

    if(event->err_msg)
	(*event->err_msg)(event, mt, file, func, line, message);
	
    /*printf("errmsg mt: %i, file: %s, func: %s line: %i.\n%s\n",
	mt, file, func, line, message);*/

    free(file);
    free(func);
    free(message);
}

/* called on a <event type="ping"  */
static void ngeclient_handle_ping(nge_event * event, char *tag, int chars)
{
    if(event->ping)
	(*event->ping)(event);
    /*printf("event-tag \"%s\", %i chars.\n", tag, chars);*/
}

/* called on a <event type= */
static void ngeclient_handle_event(nge_event * event, char *tag, int chars)
{
    char * type=ngeclient_get_option(tag, "type");
    
    if(strcmp(type, "service_state_change")==0)
    {
	ngeclient_handle_service_state_change(event, tag, chars);
    }
    else if (strcmp(type, "system_state_change")==0)
    {
	ngeclient_handle_system_state_change(event, tag, chars);
    }
    else if (strcmp(type, "service_output")==0)
    {
	ngeclient_handle_service_output(event, tag, chars);
    }
    else if (strcmp(type, "err_msg")==0)
    {
	ngeclient_handle_err_msg(event, tag, chars);
    }
    if(strcmp(type, "initial_service_state")==0)
    {
	ngeclient_handle_initial_service_state(event, tag, chars);
    }
    else if (strcmp(type, "initial_system_state")==0)
    {
	ngeclient_handle_initial_system_state(event, tag, chars);
    }    
    else if (strcmp(type, "ping")==0)
    {
	ngeclient_handle_ping(event, tag, chars);
    }
    else
    {
	ngeclient_error="Unknown event-tag.";
	/*fprintf(stderr, "Unknown event-tag \"%s\"\n", type);*/
    }
    
    free(type);
}

/* called on a <connect protocol_version="1" initng_version="initng-123"/> */
static void ngeclient_handle_connect(nge_event * event, char *tag, int chars)
{
    char *initng_version=ngeclient_get_option(tag, "initng_version");
    int protocol_version=ngeclient_get_int(tag, "protocol_version");
    if(protocol_version != NGE_VERSION)
    {
	ngeclient_error="NGE protocol version missmatch!";
	/*fprintf(stderr, "NGE protocol version missmatch!, (%i, %i) ABORT!\n", protocol_version, NGE_VERSION);*/
	free(initng_version);
	return;
    }	
    if(event->connect)
	(*event->connect)(event, protocol_version, initng_version);
    /*printf("Connected to initng version: \"%s\" \n", initng_version);*/
    free(initng_version);
}

/* called when a <dissconnect> hook apeares */
static void ngeclient_handle_dissconnect(nge_event * event, char *tag, int chars)
{
    if(event->dissconnect)
	(*event->dissconnect)(event);
    /*printf("dissconnect-tag \"%s\", %i chars.\n", tag, chars);*/
}

/* called for every tag nge fetches */
static void ngeclient_handle_tag(nge_event * event, char *tag, int chars)
{
    /* if this is a <? comment > tag */
    if(tag[1]=='?')
	return;

    /* if this is a "<event " tag */
    if(strncmp(&tag[1], "event ", 6)==0)
    {
	ngeclient_handle_event(event, tag, chars);
	return;
    }

    /* if this is a "<connect " tag */
    if(strncmp(&tag[1], "connect ", 8)==0)
    {
	ngeclient_handle_connect(event, tag, chars);
	return;
    }

    /* if this is a "<dissconnect>" tag */
    if(strncmp(&tag[1], "dissconnect>", 12)==0)
    {
	ngeclient_handle_dissconnect(event, tag, chars);
	return;
    }
	
    ngeclient_error="Unknown tag to handle.";
    /*fprintf(stderr, "Dont know how to handle tag \"%s\", %i chars.\n", tag, chars);*/
}

/*
 * ngeclient_parse_tags()
 * This functions sorts out the tags from read_buffer, and 
 * calls them one by one to ngeclient_handle_tag();
 */
void ngeclient_parse_tags(nge_event * event)
{
    assert(event);
    
    /* make sure there is data */
    if(!event->read_buffer || event->read_buffer_len < 1)
	return;

    if(event->read_buffer[0] != '<')
    {
	ngeclient_error="Expected an < char, as the first char in an xml protocol.";
	/*fprintf(stderr, "Bad xml! '%c' len: %i\n", event->read_buffer[0], strlen(event->read_buffer));*/
	return;
    }
    
    /* while there is data to read */
    while(event->read_buffer[0]!='\0')
    {
	/* skip newlines */
	while(event->read_buffer[0]=='\n')
	{
	    /* cut first char from event->read_buffer */
	    ngeclient_cut_buffert(event, 1);
	    continue;
	}
	
	/* count the lengt of this tag */
	int chars=0;
	while(event->read_buffer[chars])
	{
	    /* if found "/>" This meens tag-out */
	    if(event->read_buffer[chars]=='/' && event->read_buffer[chars+1]=='>')
	    {
		chars+=2;
		break;
	    }
	    
	    /* if found "</" This is also tag-out */
	    if(event->read_buffer[chars]=='<' && event->read_buffer[chars+1]=='/')
	    {
		chars+=2;
		
		/* skip to the end char of tag */
		while(event->read_buffer[chars] && event->read_buffer[chars] != '>')
		    chars++;
		break;
	    }
	    
	    /* increase chars of tags counter */
	    chars++;
	}
	
	
	    if(chars>0 && event->read_buffer[0]=='<')
	    {
		char *tmp=strndup(event->read_buffer, chars);
		ngeclient_handle_tag(event, tmp, chars);
		free(tmp);	
	    }
	    ngeclient_cut_buffert(event, chars);
	}
}


/*
 * This is a main loop for appications, that wont
 * care about the app to do other things.
 */
void ngeclient_exec(nge_event *event)
{    
    int got;
    
    assert(event);
    
    if(event->sock < 1)
    {
	ngeclient_error="Socket not open";
	/*fprintf(stderr, "Must call open_socket before entering ngeclient loop.\n");*/
	return;
    }
    
    /* poll_for_input will return -1 if socket closes */
    while((got=ngeclient_poll_for_input(event, 10000)>=0))
    {	
	/* if we got something */
	if(got>0 && event->read_buffer && event->read_buffer[0])
	    ngeclient_parse_tags(event);
    }
}
