
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <sys/mount.h>
#include <termios.h>
#include <stdio.h>
#include <sys/klog.h>
#include <errno.h>
#ifdef SELINUX
#include <selinux/selinux.h>
#include <selinux/get_context_list.h>
#endif
#include "initng_global.h"
#include "initng_signal.h"
#include "initng_handler.h"
#include "initng_execute.h"
#include "initng_active_db.h"
#include "initng_load_module.h"
#include "initng_plugin_callers.h"
#include "initng_toolbox.h"


#include "initng_main.h"
/*extern char **environ; */

static int local_sulogin_count = 0;
static void initng_hard(h_then t);
static void initng_main_exec_sc(h_then t);

/*
 * Poll if there are any processes left, and if not quit initng
 */
int initng_main_ready_to_quit(void)
{
    active_db_h *current = NULL;

    /* If the last process has died, quit initng */
    while_active_db(current)
    {
        /* Don't check with failed services */
        if (IS_FAILED(current))
            continue;

        /* if its down, its nothing to count on */
        if (IS_DOWN(current))
            continue;

        /*
         * if we got here,
         * its must be an non failing service, 
         * left.
         */
        return (FALSE);
    }

    return (TRUE);
}

/* this is called when there is no processes left */
void initng_main_when_out(void)
{
    if (g.i_am_init && getpid() != 1)
    {
        F_("I AM NOT INIT, THIS CANT BE HAPPENING!\n");
        sleep(3);
        return;
    }

    /* always good to do */
    sync();

    /* none of these calls should return, so the su_login on the end will be a fallback */
    switch (g.when_out)
    {
        case THEN_QUIT:
            P_(" ** Now Quiting **\n");
            initng_main_exit(0);
            break;
        case THEN_SULOGIN:
            P_(" ** Now SuLogin\n");
            /* break here leads to su_login below */
            break;
        case THEN_RESTART:
            P_(" ** Now restarting\n");
            initng_main_restart();
            break;
        case THEN_NEW_INIT:
            P_(" ** Launching new init\n");
            initng_main_new_init();
            break;
        case THEN_SC_REBOOT:
        case THEN_SC_HALT:
            initng_main_exec_sc(g.when_out);
            break;
        case THEN_REBOOT:
        case THEN_HALT:
            initng_hard(g.when_out);
            break;
    }

    /* fallback */
    initng_main_su_login();
}

static void initng_hard(h_then t)
{
    FILE *test;
    int pid;

    /* set the sys state */
    if (t == THEN_REBOOT)
        initng_main_set_sys_state(STATE_REBOOT);
    if (t == THEN_HALT)
        initng_main_set_sys_state(STATE_HALT);

    /* sync data to disk */
    sync();

    /* make sure we are root */
    if (getuid() != 0)
        return;

    /* unload all modules (plugins) found */
    initng_unload_module_unload_all();

    /* Mount readonly, youst to be extra sure this is done */
    mount("/dev/root", "/", NULL, MS_RDONLY + MS_REMOUNT, NULL);

    if (errno == EBUSY)
    {
        F_("Failed to remount / ro, EBUSY\n");
    }

    /* check so that / is mounted read only, by trying to open a file */
    if ((test = fopen("/initng_write_testfile", "w")) != NULL)
    {
        fclose(test);
        unlink("/initng_write_testfile");
        F_("/ IS NOT REMOUNTED READ-ONLY, WONT REBOOT/HALT BECAUSE THE FILE SYSTEM CAN BREAK!\n");
        return;
    }

    /* Under certain unknown circumstances, calling reboot(RB_POWER_OFF) from
       pid 1 leads to a "Kernel panic - not syncing: Attempted to kill init!".
       Workaround is to fork a child to do it. See bug #488 for details */
    if (t == THEN_REBOOT || t == THEN_HALT)
    {
        pid = fork();
        if (pid == 0)
        {
            /* child process - shut down the machine */
            if (t == THEN_REBOOT)
                reboot(RB_AUTOBOOT);
            if (t == THEN_HALT)
                reboot(RB_POWER_OFF);
            _exit(0);
        }
        else if (pid < 0)
        {
            F_("Failed to fork child process to reboot\n");
            return;
        }
        else
        {
            /* parent process waits for child to exit */
            waitpid(pid, NULL, 0);
        }
    }

    /* idle forever */
    while (1)
        sleep(1);


}

#define TRY_TIMES 2
/*
 *  Launch sulogin and wait it for finish
 */
void initng_main_su_login(void)
{
    pid_t sulogin_pid;
    int status;

#ifdef SELINUX
    if (is_selinux_enabled > 0)
    {
        security_context_t *contextlist = NULL;

        if (get_ordered_context_list("root", 0, &contextlist) > 0)
        {
            if (setexeccon(contextlist[0]) != 0)
                fprintf(stderr, "setexeccon failed\n");
            freeconary(contextlist);
        }
    }
#endif
    /* sulogin nicely 2 times */
    if (local_sulogin_count <= TRY_TIMES)
    {
        printf("This is a sulogin offer,\n"
               "you will be able to login for %i times (now %i),\n"
               "and on return initng will try continue where it was,\n"
               "if the times go out, initng will launch\n"
               "/sbin/initng-segfault on next su_login request.\n\n",
               TRY_TIMES, local_sulogin_count);
        sulogin_pid = fork();

        if (sulogin_pid == 0)
        {
            char *argv[] = { i_strdup("sulogin"), NULL };

            execve("/sbin/sulogin", argv, environ);     /* launch sulogin */

            printf("Unable to execute /sbin/sulogin!\n");
            initng_main_segfault();
            _exit(1);
        }

        if (sulogin_pid > 0)
        {
            do
            {
                sulogin_pid = waitpid(sulogin_pid, &status, WUNTRACED);
            }
            while (!WIFEXITED(status) && !WIFSIGNALED(status));

            /* increase the sulogin_count */
            local_sulogin_count++;
            return;
        }
    }

    /* else run segfault script */
    initng_main_segfault();

    /* newer get here */
    _exit(1);
}

/*
 * Start initial services
 */
void initng_main_start_extra_services(void)
{
    int i;
    int a_count = 0;            /* counts orders from argv to start */

    initng_main_set_sys_state(STATE_STARTING);
    /* check with argv which service to start initiating */
    for (i = 1; i < g.Argc; i++)
    {
        /* start all services that is +service */
        if (g.Argv[i][0] != '+')
            continue;

        /* if succeed to load this service */
        if (initng_handler_start_new_service_named(g.Argv[i] + 1))
            a_count++;
        else
            F_(" Requested service \"%s\", could not be executed!\n",
               g.Argv[i]);
    }
}

/* This is same as execve() */
void initng_main_new_init(void)
{
    int i;

    initng_main_set_sys_state(STATE_EXECVE);
    for (i = 3; i <= 1013; i++)
    {
        close(i);
    }
    if (!g.new_init || !g.new_init[0])
    {
        F_(" g.new_init is not set!\n");
        return;
    }
    P_("\n\n\n          Launching new init (%s)\n\n", g.new_init[0]);
    execve(g.new_init[0], g.new_init, environ);
}

/* execve() reboot script */
static void initng_main_exec_sc(h_then t)
{
    char **argv = NULL;
    int i;

    /* set the sys_state */
    if (t == THEN_SC_REBOOT)
        initng_main_set_sys_state(STATE_REBOOT);
    if (t == THEN_SC_HALT)
        initng_main_set_sys_state(STATE_HALT);

    /* close all open sockets, next init cant use them. */
    for (i = 3; i <= 1013; i++)
    {
        close(i);
    }

    /* allocate an argv */
    argv = (char **) i_calloc(3, sizeof(char *));

    /* set the shutdown_script path */
    argv[0] = i_strdup("/sbin/shutdown_script");

    /* set the argument to shutdown_script */
    if (t == THEN_SC_REBOOT)
        argv[1] = i_strdup("reboot");
    if (t == THEN_SC_HALT)
        argv[1] = i_strdup("poweroff");

    P_("\n\n\n          Launching %s (%s)\n", argv[0], argv[1]);
    execve(argv[0], argv, environ);
}

void initng_main_restart(void)
{
    char **argv = NULL;
    int i;

    initng_main_set_sys_state(STATE_RESTART);
    for (i = 3; i <= 1013; i++)
    {
        close(i);
    }
    argv = (char **) i_calloc(3, sizeof(char *));

    argv[0] = (char *) i_calloc(strlen(g.runlevel) + 12, sizeof(char));
    strcpy(argv[0], "runlevel=");
    strcat(argv[0], g.runlevel);
    argv[1] = NULL;
    P_("\n\n\n          R E S T A R T I N G,  (Really hot reboot)\n\n");
    execve("/sbin/initng", argv, environ);
}

/* this function sets g.sys_state, and call plug-ins that listen on its change */
void initng_main_set_sys_state(h_sys_state state)
{
    /* don't set a state that is */
    if (state == g.sys_state)
        return;

    D_("set_sys_state(): %% Setting state to: %i %% \n", state);
    g.sys_state = state;

    /*
     * execute all functions in modules that want to
     * be executed when system state change occurs.
     */

    initng_plugin_callers_load_module_system_changed(state);
    return;
}


void initng_main_exit(int i)
{
    D_("exit_initng();\n");

    /* First set the system state to exit */
    initng_main_set_sys_state(STATE_EXIT);

    /* Free all global variables */
    initng_global_free();

    /* Then, unload all modules */
    initng_unload_module_unload_all();

    /* And exit with return code */
    exit(i);
}

/*
 * This is called on segfault,
 * DON'T call any other function in initng from here, not even P_, F_, W_!
 * Also, do NOT call any function that allocates memory (using malloc),
 * since the SIGSEGV might be due to a corrupted malloc arena
 */
void initng_main_segfault(void)
{
    char *argv[3];

    /* if this is not init */
    if (!g.i_am_init)
    {
        /* just quit */
        _exit(99);
    }

#ifdef DEBUG
    /* if we exit the program before sulogin, we wont be able to gdb initng */
    /*su_login(); */
    printf("Initng segfaulted, will wait in 20 seconds for you to start a gdb, before execve(/sbin/initng-segfault);\n");
    while (1)
        sleep(20);
#endif

    /*
     * IF we close sockets, it gets problems, when reloading after a segfault.
     int i
     for (i = 3; i <= 1013; i++)
     {
     close(i);
     }
     */

    argv[0] = (char*)"/sbin/initng-segfault";
    argv[1] = NULL;

    printf("Launching /sbin/initng-segfault, this is pid: %i\n", getpid());
    if (execve("/sbin/initng-segfault", argv, environ) == -1)
    {
        printf("/sbin/initng-segfault did not exist, will die!\n");
    }
    _exit(99);
}

void initng_main_set_runlevel(const char *runlevel)
{
    /* first free the old_runlevel */
    if (g.old_runlevel)
    {
        free(g.old_runlevel);
        g.old_runlevel = NULL;
    }

    /* then move the last new, to the old */
    if (g.runlevel)
    {
        g.old_runlevel = g.runlevel;
        g.runlevel = NULL;
    }

    /* and set the new */
    g.runlevel = i_strdup(runlevel);
}
